import json, timeit, gc
from collections import defaultdict, namedtuple


# DLL
def dll_init():
    head = []
    head[:] = [head, head, None]
    return head

def dll_append(head, value):
    last = head[0]
    node = [last, head, value]
    last[1] = head[0] = node
    return node

def dll_remove(node):
    prev_link, next_link, _ = node
    prev_link[1] = next_link
    next_link[0] = prev_link
    return node

def dll_insert_before(succedent, value):
    # Create a node to host value, and insert BEFORE the succedent
    node = [succedent[0], succedent, value]
    succedent[0][1] = node
    succedent[0] = node
    return node

def dll_insert_after(precedent, value):
    # Create a node to host value, and insert AFTER the precedent
    node = [precedent, precedent[1], value]
    precedent[1][0] = node
    precedent[1] = node
    return node

def dll_iter(head):
    curr = head[1]              # start at the first node
    while curr is not head:
        yield curr[2]           # yield the curr[KEY]
        curr = curr[1]          # move to next node

# LFUCache2


class LFUCache2(object):
    def __init__(self, capacity, *args):
        self.capacity = capacity
        self.bucket_lut = defaultdict(dll_init)
        self.cache = dict()

    def remove_node(self, node):
        # Remove the node from the cache and also the bucket
        _, _, (key, value, counter) = node
        dll_remove(node)
        self.cache.pop(key)
        # clean up the freq_head if it is empty
        bucket = self.bucket_lut[counter]
        if bucket[1] is bucket:
            self.bucket_lut.pop(counter)

    def add_node(self, data):
        # Create a node to host the data, add it to cache and
        # append to the bucket.
        bucket = self.bucket_lut[data[2]]
        node = dll_append(bucket, data)
        self.cache[data[0]] = node

    def set(self, key, value):
        # special case for capacity <= 0
        if self.capacity <= 0:
            return

        # Does the key exist in the cache?
        node = self.cache.get(key)
        if node:
            # Update the value and counter
            counter = node[2][2]
            self.remove_node(node)
            self.add_node((key, value, counter + 1))
            return

        if len(self.cache) >= self.capacity:
            # Remove the least used, least recently accessed
            min_counter = min(self.bucket_lut.keys())
            bucket = self.bucket_lut[min_counter]
            self.remove_node(bucket[1])

        self.add_node((key, value, 1))

    def get(self, key):
        node = self.cache.get(key)
        if node is None:
            return -1

        key, value, counter = node[2]
        self.remove_node(node)
        self.add_node((key, value, counter + 1))
        return value


# LFUCache3

# head points to the doubly linked list of Node
Bucket = namedtuple('Bucket', ['counter', 'head'])
# item caches a reference to the bucket_node for quick accss the next bucket
Item = namedtuple('Item',['key', 'value', 'bucket_node'])

class LFUCache3(object):
    def __init__(self, capacity, *args):
        self.capacity = capacity
        self.bucket_head = dll_init()
        self.cache = dict()

    def remove_node(self, node):
        item = node[2]
        self.cache.pop(item.key)   # remove from cache
        dll_remove(node)           # remove node from the bucket

        bucket = item.bucket_node[2]
        if bucket.head[1] is bucket.head:
            # remove the bucket if empty
            dll_remove(item.bucket_node)

    def add_node(self, key, value, original_bucket_node):
        '''Add the (key, value) content pulled from orginal_bucket_node
        to a new bucket'''
        counter = 0 if original_bucket_node is self.bucket_head \
            else (original_bucket_node[2].counter)
        next_bucket_node = original_bucket_node[1]

        if next_bucket_node is self.bucket_head:
            # No bucket(counter + k) exists, append a new bucket(counter + 1)
            bucket = Bucket(counter + 1, dll_init())
            bucket_node = dll_append(self.bucket_head, bucket)
        elif next_bucket_node[2].counter != counter + 1:
            # bucket(counter + k) exist, insert bucket(counter + 1) BEFORE next_bucket_node
            bucket = Bucket(counter + 1, dll_init())
            bucket_node = dll_insert_before(next_bucket_node, bucket)
        else:
            # bucket(counter + 1) exists, use it
            bucket = next_bucket_node[2]
            bucket_node = next_bucket_node

        # Create the item, append it to the bucket and add to the cache.
        item = Item(key, value, bucket_node)
        self.cache[key] = dll_append(bucket.head, item)

    def set(self, key, value):
        # special case for capacity <= 0
        if self.capacity <= 0:
            return

        # Does the key exist in the cache?
        node = self.cache.get(key)
        if node:
            item = node[2]
            self.remove_node(node)
            self.add_node(item.key, value, item.bucket_node)
            return

        if len(self.cache) >= self.capacity:
            # Apply LRFU alogrithm here!
            bucket = self.bucket_head[1][2]
            self.remove_node(bucket.head[1])

        self.add_node(key, value, self.bucket_head)

    def get(self, key):
        node = self.cache.get(key)
        if node is None:
            return -1

        item = node[2]
        self.remove_node(node)
        self.add_node(item.key, item.value, item.bucket_node)
        return item.value


def setup():
    with open('lfu-cache-test-fixture.json') as f:
        fixture = json.load(f)
        cache = LFUCache2(2048)  # or LFUCache3
        return cache, fixture


def benchmark(cache, fixture):
    lut = {
        'set': cache.set,
        'get': cache.get
    }
    for method, args in zip(*fixture):
        lut[method](*args)
    gc.collect()

if __name__ == "__main__":
    print(timeit.timeit('benchmark(*setup())', setup='from __main__ import setup, benchmark', number=1000))
